package pl8.extractor;

import java.io.*;
import java.util.*;

import ar.com.hjg.pngj.ImageInfo;
import ar.com.hjg.pngj.PngWriter;

public class PL8Extractor {

	static enum Type { ORTHOGONAL, ISOMETRIC, RLE_ENCODED, UNKNOWN };

	private static final boolean writeSingleTiles = false; // write every tile as single png
	private static final boolean writeTileset = true; // write all tiles of a pl8 to a single png
	private static final boolean allPalettes = false; // write the same pl8 multiple times with different palettes

	private static final String folder = "PL8/"; // folder name incl. backslash

	public static void main (String [] args) {
		try {
			extractAll();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public static void extractAll() throws Exception {
		List<PL8Entry> entries = CSVParser.parseCSV("/res/pl8_palette_type.csv");
		
		Map<String, Integer[]> palettes = new HashMap<String, Integer[]>();
		
		int counter = 1;
		
		for (PL8Entry entry : entries) {
			boolean is_pl8 = true;
			
			String pl8filename = folder + entry.filename;
			
			System.out.print((counter++) + "/" + entries.size() + ": ");
			
			// check if file exists
		  	File pl8file = new File(pl8filename);
		  	if (!pl8file.exists()) {
		  		pl8file = new File(pl8filename + ".PL8");
		  		if (! pl8file.exists()) {
		  			System.out.println("Skipping " + entry.filename + ".PL8");
		  			continue;
		  		}
		  		System.out.print("Extracting " + entry.filename + ".PL8...");
		  		pl8filename += ".PL8";
		  	}
		  	
		  	if (!palettes.containsKey(entry.palette)) {
		  		Integer pal[] = loadPalette(folder + entry.palette + ".256");
			  	if (pal != null) {
			  		palettes.put(entry.palette, pal);
			  	}
		  	}
		  	
		  	pl8filename = pl8file.getAbsolutePath();
		  	
		  	MyStream stream = new MyStream(pl8filename);
		  	
			Tile[] tiles;
			int numTiles = 0;
			
			if (is_pl8) {
				
			    stream.readShort(); // unknown, no effect
			    numTiles = stream.readShort();
			    stream.readInt(); // skip unknown bytes
			    
			    tiles = new Tile[numTiles];
				
			    for (int i = 0; i < tiles.length; i++) {
			    	int width  = stream.readShort();
			    	int height = stream.readShort();
			    	int offset = stream.readInt();
			    	
			    	int x = stream.readShort();
			    	int y = stream.readShort();
			    	
			    	int extraType = stream.readByte();
			    	int extraRows = stream.readByte();
			    	
			    	stream.readByte(); // always 0
			    	stream.readByte(); // always 0 except in FONT3C2.PL8 and FNTL2_22.PL8
			    	
			    	tiles[i] = new Tile(width, height, offset, x, y, extraType, extraRows);
				}
			  
			} else {
				
				/*
					There are 40 map slots in L2_MAPS.DAT but only 24 are used
					Each map has six 64*64 bitmaps and one 65*129 bitmap
				*/
				
				tiles = new Tile[24 * 7];
				
				int width[]  = new int[tiles.length];
				int height[] = new int[tiles.length];
				int offset[] = new int[tiles.length];
				
				for (int i = 0; i < tiles.length; i++) {
					width[i]  = 64;
					height[i] = 64;
					offset[i] = (i/7) * 32961 + (i % 7) * 4096;
				}
				for (int i = 6; i < tiles.length; i+=7) {
					width[i]  = 65;
					height[i] = 129;
				}
				
				for (int i = 0; i < tiles.length; i++) {
					tiles[i] = new Tile(width[i], height[i], offset[i]);
				}
			}
			
			ITileset tileset = null;
			
			if (writeTileset) {
				tileset = new Tileset();
				//tileset = new TilesetFixed(32, 32);
			}
			
			Integer palette[] = palettes.get(entry.palette);
			
			if (palette == null) {
				System.out.print("palette not found, using default...");
				palette = getDefaultPalette();
			}
			
			for (int i = 0; i < tiles.length; i++) {
				
				int width = tiles[i].getWidth();
				int height = tiles[i].getHeight();
				int offset = tiles[i].getOffset();
				int x = tiles[i].getX();
				int y = tiles[i].getY();
				int extraType = tiles[i].getExtraType();
				int extraRows = tiles[i].getExtraRows();
				
				if (width == 0 || height == 0) {
					continue;
				}
				
				byte bmp[] = new byte[height * width];
				
				// tile contains the array for the PNG in RGBA format
				int tile[][] = new int[height][width * 4];
				int extra[][] = null;
				
				stream.seek(offset);
				
				if (entry.type == Type.ORTHOGONAL) {
					
					// read data into byte array
					stream.read(bmp);
					
					for (int h = 0; h < height; h++) {
						for (int w = 0; w < width; w++) {
							int index = (bmp[h*width + w] & 0xFF) * 3;
							
							// check if transparent
							if (index != 0 || !is_pl8) {
								tile[h][w*4]     = palette[index + 2];
								tile[h][w*4 + 1] = palette[index + 1];
								tile[h][w*4 + 2] = palette[index];
								tile[h][w*4 + 3] = 0xFF;
							}
						}
					}
					
				} else if (entry.type == Type.ISOMETRIC) {
					
					// upper part of the tile
					for (int h = 0; h < 15; h++) {
						for (int w = 0; w < (h*4)+2; w++) {
							int b = stream.readByte();
							
							int ww = (14-h)*2 + w;
							bmp[h*58 + ww] = (byte) b;
							
							ww *= 4;
							int index = b * 3;
							tile[h][ww]     = palette[index + 2];
							tile[h][ww + 1] = palette[index + 1];
							tile[h][ww + 2] = palette[index];
							tile[h][ww + 3] = 0xFF;
						}
					}
					
					// lower part of the tile
					for (int h = 0; h < 15; h++) {
						for (int w = 0; w<((14-h)*4)+2; w++) {
							int b = stream.readByte();
							
							int ww = h*2 + w;
							int hh = h + 15;
							int index = b * 3;
							bmp[hh*58 + ww] = (byte) b;
							
							ww *= 4;
							tile[hh][ww]     = palette[index + 2];
							tile[hh][ww + 1] = palette[index + 1];
							tile[hh][ww + 2] = palette[index];
							tile[hh][ww + 3] = 0xFF;
						}
					}
					
					if (extraType != 1) {
					
						// read extra rows
						int leftOffset = 0;
						int rightOffset = width;
						
						if (extraType == 3) { // left only
							rightOffset = width/2 + 1;
						} else if (extraType == 4) { // right only
							leftOffset = width/2 - 1;
						}
						
						int halfHeight = height>>1;
						int halfWidth = width>>1;
						extra = new int[extraRows + halfHeight][width * 4];
						
						for (int h = 0; h < extraRows; h++) {
							
							byte[] row = new byte[width];
							for (int z = leftOffset; z < rightOffset; z++) {
								row[z] = (byte) stream.readByte();
							}
							
							for (int w = leftOffset; w < rightOffset; w++) {
								int yPos = extraRows - h;
								
								if (w <= halfWidth) {
									yPos += (halfHeight-1) - (w/2);
								} else {
									yPos += (w/2) - (halfHeight-1);
								}
								
								int index = (row[w] & 0xFF) * 3;
								
								if (index != 0) {
									int ww = w*4;
									extra[yPos][ww]     = palette[index + 2];
									extra[yPos][ww + 1] = palette[index + 1];
									extra[yPos][ww + 2] = palette[index];
									extra[yPos][ww + 3] = 0xFF;
								}
							}
						}
					}
					
				} else if (entry.type == Type.RLE_ENCODED) {
					
					for (int h = 0; h < height; h++) {
						
						int w = 0;
						while (w < width) {
							
							int numOpqauePixels = stream.readByte();
							
							if (numOpqauePixels == 0) {
								
								int numTransparentPixels = stream.readByte();
								w += numTransparentPixels;
								
							} else {
								
								for (int z = 0; z < numOpqauePixels; z++) {
									int index = stream.readByte() * 3;
									
									tile[h][w*4]     = palette[index + 2];
									tile[h][w*4 + 1] = palette[index + 1];
									tile[h][w*4 + 2] = palette[index];
									tile[h][w*4 + 3] = 0xFF;
									
									w++;
								}
							}
						}
					}
					
				} else {
					System.out.println("Unknown type!");
				}
				
				//tiles[i].setBitmap(bmp);
				
				if (writeTileset) {
					
					if (entry.type == Type.ISOMETRIC) {
						int[][] newTile = new int[64][64 * 4];
						
						for (int j = 0; j < 30; j++) {
							for (int k = 0; k < 58 * 4; k++) {
								newTile[j+34][k] = tile[j][k];
							}
						}
						
						if (extraType != 1 && extraType != 0) {
							for (int j = 0; j < extra.length; j++) {
								for (int k = 0; k < 58*4; k+=4) {
									if (extra[j][k+3] != 0) {
										for (int m = 0; m < 4; m++) {
											newTile[j + 33-extraRows][k+m] = extra[j][k+m];
										}
									}
								}
							}
							
							//tileset.addTile(x, y, tile);
							//tileset.addTile(x, y-extraRows-1, extra);
						}
						
						// use if fixed tileset
						//tileset.addTile(x, y, newTile);
						
						tileset.addTile(x, y-34, newTile);
						
					} else {
						
						tileset.addTile(x, y, tile);
					}
				}
				
				if (writeSingleTiles) {
					// Create new output file
					String outFilename = pl8filename + "." + i + ".PNG";
					
				  	FileOutputStream fos;
				  	try {
				  		fos = new FileOutputStream(new File(outFilename));
				  	} catch (FileNotFoundException ex) {
				  		System.out.println("\nCouldn't open " + outFilename + "!");
				  		continue;
				  	}
				  	
				  	// Write PNG
				 	BufferedOutputStream outputStream = new BufferedOutputStream(fos);
				 	ImageInfo imgInfo = new ImageInfo(width, height, 8, true);
				 	PngWriter writer = new PngWriter(outputStream, imgInfo);
				 	
				 	for (int h = 0; h < height; h++) {
				 		writer.writeRowInt(tile[h]);
				 	}
				 	
				 	writer.end();
				}
			
			}
			
			if (writeTileset) {
				pl8filename = pl8filename.substring(0, pl8filename.length()-4); // remove .PL8 extension
				
				tileset.writeTileset(pl8filename + ".PNG");
				
				/*
				switch (entry.type) {
				case ORTHOGONAL:
					pl8filename += ".BMP";
					break;
				case ISOMETRIC:
					pl8filename += ".ISO";
					break;
				case RLE_ENCODED:
					pl8filename += ".RLE";
					break;
				default:
					break;
				}
				
				if (entry.palette.length() > 0) {
					tileset.writeTileset(pl8filename + "." + entry.palette + ".PNG");
				} else {
					tileset.writeTileset(pl8filename + ".PNG");
				}
				*/
			}
			
			System.out.println("done!");
		  	stream.close();
		}
	}

	public static void extractSingle() {
		
		String palName = "BASE01"; // palette filename without .256 extension
		Type type = Type.ISOMETRIC; // type of the PL8 file(s)

		String[] filenames = { // the PL8 files to extract
			"Castle1a"
		};
		
		String[] palNames = { // use these palettes if (allPalettes == true)
			"ARMITEMS", "ARMOURY", "BACKGRND", "BASE01", "CASTLE1", "CAS_BACK", "CUSTOM", "GATEWAY",
			"GRTNOBLE", "LORDS2", "MERCHANT", "SPRITE01", "SPRITE1A", "T32_BAT1", "T32_STN1", "TREASURY",
			"SKIRMISH"
		};
		
		List<Integer[]> palettes = new ArrayList<Integer[]>();
		
		for (String pl8filename : filenames) {
		
			pl8filename = folder + pl8filename;
			
			try {
		  	
		  	// check if file exists
		  	File pl8file = new File(pl8filename);
		  	if (!pl8file.exists()) {
		  		pl8file = new File(pl8filename + ".PL8");
		  		if (! pl8file.exists()) {
		  			System.out.println("File " + pl8filename + " does not exist!");
		  			System.exit(1);
		  		}
		  		pl8filename += ".PL8";
		  	}
		  	
		  	boolean is_pl8;
		  	
		  	// L2_MAPS.DAT contains the actual maps of the game
		  	if (pl8filename.equals(folder + "L2_MAPS.DAT")) {
		  		is_pl8 = false;
		  	} else {
		  		is_pl8 = true;
		  	}
		  	
		  	palettes.clear();
		  	
		  	if (allPalettes) {
		  		for (String p : palNames) {
		  			Integer pal[] = loadPalette(folder + p + ".256");
				  	if (pal != null) {
				  		palettes.add(pal);
				  	}
		  		}
		  	} else {
			  	// check if the PL8 has its own color palette
			  	if (palName.length() == 0) {
			  		palName = pl8filename.substring(0, pl8filename.length()-4);
			  	}
			  	
			  	Integer pal[] = loadPalette(folder + palName + ".256");
			  	if (pal != null) {
			  		palettes.add(pal);
			  	}
		  	}
		  	
		  	// use default palette if no palette was specified
			if (palettes.isEmpty()) {
				palettes.add(getDefaultPalette());
			}
		  	
		  	pl8filename = pl8file.getAbsolutePath();
		  	
		  	//MyRandomAccessFile stream = new MyRandomAccessFile(pl8filename, "r");
		    MyStream stream = new MyStream(pl8filename);
		  	
			Tile[] tiles;
			int numTiles = 0;
			
			if (is_pl8) {
				
			    stream.readShort(); // unknown, no effect
			    numTiles = stream.readShort();
			    stream.readInt(); // skip unknown bytes
			    
			    tiles = new Tile[numTiles];
				
			    for (int i = 0; i < tiles.length; i++) {
			    	int width  = stream.readShort();
			    	int height = stream.readShort();
			    	int offset = stream.readInt();
			    	
			    	int x = stream.readShort();
			    	int y = stream.readShort();
			    	
			    	int extraType = stream.readByte();
			    	int extraRows = stream.readByte();
			    	
			    	stream.readByte(); // always 0
			    	stream.readByte(); // always 0 except in FONT3C2.PL8 and FNTL2_22.PL8
			    	
			    	tiles[i] = new Tile(width, height, offset, x, y, extraType, extraRows);
				}
			  
			} else {
				
				/*
					There are 40 map slots in L2_MAPS.DAT but only 24 are used
					Each map has six 64*64 bitmaps and one 65*129 bitmap
				*/
				
				tiles = new Tile[24 * 7];
				
				int width[]  = new int[tiles.length];
				int height[] = new int[tiles.length];
				int offset[] = new int[tiles.length];
				
				for (int i = 0; i < tiles.length; i++) {
					width[i]  = 64;
					height[i] = 64;
					offset[i] = (i/7) * 32961 + (i % 7) * 4096;
				}
				for (int i = 6; i < tiles.length; i+=7) {
					width[i]  = 65;
					height[i] = 129;
				}
				
				for (int i = 0; i < tiles.length; i++) {
					tiles[i] = new Tile(width[i], height[i], offset[i]);
				}
			}
			
			ITileset tileset = null;
			
			if (writeTileset) {
				tileset = new Tileset();
				//tileset = new TilesetFixed(32, 32);
			}
			
			for (int palCounter = 0; palCounter < palettes.size(); palCounter++) {
				Integer palette[] = palettes.get(palCounter);
				
				for (int i = 0; i < tiles.length; i++) {
					
					int width = tiles[i].getWidth();
					int height = tiles[i].getHeight();
					int offset = tiles[i].getOffset();
					int x = tiles[i].getX();
					int y = tiles[i].getY();
					int extraType = tiles[i].getExtraType();
					int extraRows = tiles[i].getExtraRows();
					
					if (width == 0 || height == 0) {
						continue;
					}
					
					byte bmp[] = new byte[height * width];
					
					// tile contains the array for the PNG in RGBA format
					int tile[][] = new int[height][width * 4];
					int extra[][] = null;
					
					stream.seek(offset);
					
					if (type == Type.ORTHOGONAL) {
						
						// read data into byte array
						stream.read(bmp);
						
						for (int h = 0; h < height; h++) {
							for (int w = 0; w < width; w++) {
								int index = (bmp[h*width + w] & 0xFF) * 3;
								
								// check if transparent
								if (index != 0 || !is_pl8) {
									tile[h][w*4]     = palette[index + 2];
									tile[h][w*4 + 1] = palette[index + 1];
									tile[h][w*4 + 2] = palette[index];
									tile[h][w*4 + 3] = 0xFF;
								}
							}
						}
						
					} else if (type == Type.ISOMETRIC) {
						
						// upper part of the tile
						for (int h = 0; h < 15; h++) {
							for (int w = 0; w < (h*4)+2; w++) {
								int b = stream.readByte();
								
								int ww = (14-h)*2 + w;
								bmp[h*58 + ww] = (byte) b;
								
								ww *= 4;
								int index = b * 3;
								tile[h][ww]     = palette[index + 2];
								tile[h][ww + 1] = palette[index + 1];
								tile[h][ww + 2] = palette[index];
								tile[h][ww + 3] = 0xFF;
							}
						}
						
						// lower part of the tile
						for (int h = 0; h < 15; h++) {
							for (int w = 0; w<((14-h)*4)+2; w++) {
								int b = stream.readByte();
								
								int ww = h*2 + w;
								int hh = h + 15;
								int index = b * 3;
								bmp[hh*58 + ww] = (byte) b;
								
								ww *= 4;
								tile[hh][ww]     = palette[index + 2];
								tile[hh][ww + 1] = palette[index + 1];
								tile[hh][ww + 2] = palette[index];
								tile[hh][ww + 3] = 0xFF;
							}
						}
						
						if (extraType != 1) {
						
							// read extra rows
							int leftOffset = 0;
							int rightOffset = width;
							
							if (extraType == 3) { // left only
								rightOffset = width/2 + 1;
							} else if (extraType == 4) { // right only
								leftOffset = width/2 - 1;
							}
							
							int halfHeight = height>>1;
							int halfWidth = width>>1;
							extra = new int[extraRows + halfHeight][width * 4];
							
							for (int h = 0; h < extraRows; h++) {
								
								byte[] row = new byte[width];
								for (int z = leftOffset; z < rightOffset; z++) {
									row[z] = (byte) stream.readByte();
								}
								
								for (int w = leftOffset; w < rightOffset; w++) {
									int yPos = extraRows - h;
									
									if (w <= halfWidth) {
										yPos += (halfHeight-1) - (w/2);
									} else {
										yPos += (w/2) - (halfHeight-1);
									}
									
									int index = (row[w] & 0xFF) * 3;
									
									if (index != 0) {
										int ww = w*4;
										extra[yPos][ww]     = palette[index + 2];
										extra[yPos][ww + 1] = palette[index + 1];
										extra[yPos][ww + 2] = palette[index];
										extra[yPos][ww + 3] = 0xFF;
									}
								}
							}
						}
						
					} else if (type == Type.RLE_ENCODED) {	
						
						for (int h = 0; h < height; h++) {
							
							int w = 0;
							while (w < width) {
								
								int numOpqauePixels = stream.readByte();
								
								if (numOpqauePixels == 0) {
									
									int numTransparentPixels = stream.readByte();
									w += numTransparentPixels;
									
								} else {
									
									for (int z = 0; z < numOpqauePixels; z++) {
										int index = stream.readByte() * 3;
										
										tile[h][w*4]     = palette[index + 2];
										tile[h][w*4 + 1] = palette[index + 1];
										tile[h][w*4 + 2] = palette[index];
										tile[h][w*4 + 3] = 0xFF;
										
										w++;
									}
								}
							}
						}
						
					} else {
						System.out.println("Unknown type!");
					}
					
					//tiles[i].setBitmap(bmp);
					
					if (writeTileset) {
						
						if (type == Type.ISOMETRIC) {
							int[][] newTile = new int[64][64 * 4];
							
							for (int j = 0; j < 30; j++) {
								for (int k = 0; k < 58 * 4; k++) {
									newTile[j+34][k] = tile[j][k];
								}
							}
							
							if (extraType != 1 && extraType != 0) {
								for (int j = 0; j < extra.length; j++) {
									for (int k = 0; k < 58*4; k+=4) {
										if (extra[j][k+3] != 0) {
											for (int m = 0; m < 4; m++) {
												newTile[j + 33-extraRows][k+m] = extra[j][k+m];
											}
										}
									}
								}
								
								//tileset.addTile(x, y, tile);
								//tileset.addTile(x, y-extraRows-1, extra);
							}
							
							// use if fixed tileset
							//tileset.addTile(x, y, newTile);
							
							tileset.addTile(x, y-34, newTile);
							
						} else {
							
							tileset.addTile(x, y, tile);
						}
					}
					
					if (writeSingleTiles) {
						// Create new output file
						String outFilename = pl8filename + "." + i + ".PNG";
						
					  	FileOutputStream fos;
					  	try {
					  		fos = new FileOutputStream(new File(outFilename));
					  	} catch (FileNotFoundException ex) {
					  		System.out.println("\nCouldn't open " + outFilename + "!");
					  		continue;
					  	}
					  	
					  	// Write PNG
				  		BufferedOutputStream outputStream = new BufferedOutputStream(fos);
				  		ImageInfo imgInfo = new ImageInfo(width, height, 8, true);
				  		PngWriter writer = new PngWriter(outputStream, imgInfo);
				  		
				  		for (int h = 0; h < height; h++) {
				  			writer.writeRowInt(tile[h]);
				  		}
				  		
				  		writer.end();
					}
				}
				
				if (writeTileset) {
					pl8filename = pl8filename.substring(0, pl8filename.length()-4); // remove .PL8 extension
					
					switch (type) {
					case ORTHOGONAL:
						pl8filename += ".BMP";
						break;
					case ISOMETRIC:
						pl8filename += ".ISO";
						break;
					case RLE_ENCODED:
						pl8filename += ".RLE";
						break;
					default:
						break;
					}
					
					if (allPalettes) {
						// TileName.TileType.PaletteName.PNG
						tileset.writeTileset(pl8filename + "." + palNames[palCounter] + ".PNG");
					} else if (palName.length() > 0) {
						tileset.writeTileset(pl8filename + "." + palName + ".PNG");
					} else {
						tileset.writeTileset(pl8filename + ".PNG");
					}
				}
			
			} // end pals
			
			System.out.println("Done!");
		  	
		  	stream.close();
		  	
			} catch (Exception e) {
				e.printStackTrace();
			} 
		}
	}

	// loads a .256 palette
	private static Integer[] loadPalette(String palName) throws Exception {
	  	
		try {
			MyStream dis = new MyStream(palName);
			    
			Integer[] palette = new Integer[256*3];
			for (int i = 0; i < 256; i++) {
			    palette[i*3 + 2] = dis.readByte() << 2;
			    palette[i*3 + 1] = dis.readByte() << 2;
			    palette[i*3]     = dis.readByte() << 2;
			}
			return palette;
		
		} catch (IOException e) {
			return null;
		}
	}

	// returns the default MS Bitmap palette
	private static Integer[] getDefaultPalette() {
		Integer palette[] = {
				0,0,0,0,0,128,0,128,0,0,128,128,128,0,0,128,0,128,128,128,0,192,192,192,192,220,192,200,208,212,0,32,64,0,32,96,0,32,128,0,32,160,0,32,192,0,32,224,0,
				64,0,0,64,32,0,64,64,0,64,96,0,64,128,0,64,160,0,64,192,0,64,224,0,96,0,0,96,32,0,96,64,0,96,96,0,96,128,0,96,160,0,96,192,0,96,224,0,128,0,0,128,32,0,
				128,64,0,128,96,0,128,128,0,128,160,0,128,192,0,128,224,0,160,0,0,160,32,0,160,64,0,160,96,0,160,128,0,160,160,0,160,192,0,160,224,0,192,0,0,192,32,0,
				192,64,0,192,96,0,192,128,0,192,160,0,192,192,0,192,224,0,224,0,0,224,32,0,224,64,0,224,96,0,224,128,0,224,160,0,224,192,0,224,224,64,0,0,64,0,32,64,0,
				64,64,0,96,64,0,128,64,0,160,64,0,192,64,0,224,64,32,0,64,32,32,64,32,64,64,32,96,64,32,128,64,32,160,64,32,192,64,32,224,64,64,0,64,64,32,64,64,64,64,
				64,96,64,64,128,64,64,160,64,64,192,64,64,224,64,96,0,64,96,32,64,96,64,64,96,96,64,96,128,64,96,160,64,96,192,64,96,224,64,128,0,64,128,32,64,128,64,
				64,128,96,64,128,128,64,128,160,64,128,192,64,128,224,64,160,0,64,160,32,64,160,64,64,160,96,64,160,128,64,160,160,64,160,192,64,160,224,64,192,0,64,
				192,32,64,192,64,64,192,96,64,192,128,64,192,160,64,192,192,64,192,224,64,224,0,64,224,32,64,224,64,64,224,96,64,224,128,64,224,160,64,224,192,64,224,
				224,128,0,0,128,0,32,128,0,64,128,0,96,128,0,128,128,0,160,128,0,192,128,0,224,128,32,0,128,32,32,128,32,64,128,32,96,128,32,128,128,32,160,128,32,192,
				128,32,224,128,64,0,128,64,32,128,64,64,128,64,96,128,64,128,128,64,160,128,64,192,128,64,224,128,96,0,128,96,32,128,96,64,128,96,96,128,96,128,128,96,
				160,128,96,192,128,96,224,128,128,0,128,128,32,128,128,64,128,128,96,128,128,128,128,128,160,128,128,192,128,128,224,128,160,0,128,160,32,128,160,64,
				128,160,96,128,160,128,128,160,160,128,160,192,128,160,224,128,192,0,128,192,32,128,192,64,128,192,96,128,192,128,128,192,160,128,192,192,128,192,224,
				128,224,0,128,224,32,128,224,64,128,224,96,128,224,128,128,224,160,128,224,192,128,224,224,192,0,0,192,0,32,192,0,64,192,0,96,192,0,128,192,0,160,192,
				0,192,192,0,224,192,32,0,192,32,32,192,32,64,192,32,96,192,32,128,192,32,160,192,32,192,192,32,224,192,64,0,192,64,32,192,64,64,192,64,96,192,64,128,
				192,64,160,192,64,192,192,64,224,192,96,0,192,96,32,192,96,64,192,96,96,192,96,128,192,96,160,192,96,192,192,96,224,192,128,0,192,128,32,192,128,64,
				192,128,96,192,128,128,192,128,160,192,128,192,192,128,224,192,160,0,192,160,32,192,160,64,192,160,96,192,160,128,192,160,160,192,160,192,192,160,224,
				192,192,0,192,192,32,192,192,64,192,192,96,192,192,128,192,192,160,240,251,255,164,160,160,128,128,128,0,0,255,0,255,0,0,255,255,255,0,0,255,0,255,255,
				255,0,255,255,255
		};
		return palette;
	}
}